import os
import pandas as pd
import xarray as xr
import numpy as np
from scipy.spatial import cKDTree
from datetime import date, timedelta, datetime
from mpi4py import MPI
import json
import variables_to_interpolate as vti

comm = MPI.COMM_WORLD
size = comm.Get_size()
rank = comm.Get_rank()  # rank of the current process

class NetCDFVertInterp:
    def __init__(self, expid, start_d, end_d, stn_lat, stn_lon):
        self.expid = expid
        self.start_d = start_d
        self.end_d = end_d
        self.stn_lat = stn_lat
        self.stn_lon = stn_lon
        self.n_neighbours = 4
        self.base_dir = f'/esarchive/exp/monarch/{self.expid}/'
        self.doms = [i for i in os.listdir(self.base_dir) if i.startswith("d")]
        self.bool_check = any( i in self.doms for i in os.listdir(self.base_dir))

    def _get_folder_path(self, variable):
        ''' Returns the folder path for a given variable '''
        if self.bool_check:
            return f'{self.base_dir}{dom}/hourly/{variable}/'
        else:
            return f'{self.base_dir}regional/hourly/{variable}/'

    def read_netcdfs(self, variables):
        """
        Lee archivos NetCDF para múltiples variables y devuelve un diccionario con un DataFrame por variable.
        """
        results = {}
        
        for variable in variables:
            print(f"Processing for: {variable}")
            df_var = pd.DataFrame()  # Crear un DataFrame vacío para cada variable
            
            folder_path = self._get_folder_path(variable)
            if os.path.isdir(folder_path):
                # Generar la lista de fechas en el rango dado
                dates = [
                    datetime.strftime(self.start_d + timedelta(days=i), '%Y%m%d') 
                    for i in range((self.end_d - self.start_d).days + 1)
                ]
                # Crear la lista de archivos NetCDF esperados
                list_of_nc_files = [f'{folder_path}{variable}-000_{i}00.nc' for i in dates]
                
                # Procesar los archivos de la variable
                df_var = self._process_variable(list_of_nc_files, variable)
                
                # Agregar el DataFrame procesado al diccionario
                results[variable] = df_var
            else:
                print(f"No se encontró la ruta para la variable '{variable}'.")
        
        return results

    def lon_lat_to_cartesian(self, lon, lat):
        """
        Convierte coordenadas de longitud y latitud en coordenadas cartesianas.
        Maneja entradas 1D o 2D asegurando compatibilidad de dimensiones.
        """
        A = 6378.137  # semieje mayor [km]
        E2 = 6.69437999014e-3  # excentricidad al cuadrado

        # Asegurarse de que 'lon' y 'lat' son compatibles
        if lon.ndim == 1 and lat.ndim == 1:
            lon, lat = np.meshgrid(lon, lat)  # Crear un grid 2D si las entradas son 1D

        # Convertir a radianes
        try:
            lon = lon.astype(float)
            lat = lat.astype(float)
            lon_rad = np.radians(lon)
            lat_rad = np.radians(lat)
        except ValueError:
            raise ValueError("Error: 'lon' o 'lat' contienen valores que no pueden convertirse a flotantes.")

        # Cálculo de las coordenadas cartesianas
        r_n = A / (np.sqrt(1 - E2 * (np.sin(lat_rad) ** 2)))
        x = r_n * np.cos(lat_rad) * np.cos(lon_rad)
        y = r_n * np.cos(lat_rad) * np.sin(lon_rad)
        z = r_n * (1 - E2) * np.sin(lat_rad)

        # Aplanar las matrices para que sean vectores 1D
        return x.flatten(), y.flatten(), z.flatten()


    def interpolate(self, source, target, dist, inds, variable):
        tmp = []
        for t in range(source["time"].shape[0]):
            w = 1.0 / dist ** 2
            interp_values = np.sum(w * source[variable][t].values.flatten()[inds], axis=1) / np.sum(w, axis=1)
            interp_values.shape = target.shape
            tmp.append(interp_values)
        return tmp

    def save_data(self, output_folder, station_name, variables):
        """Guarda los datos de cada variable en archivos CSV dentro de carpetas específicas."""
        #create output folder if it doesn't exist
        directory_path = os.path.join(output_folder, self.expid, str(self.start_d.year))
        os.makedirs(directory_path, exist_ok=True)
        data_dict = self.read_netcdfs(variables)

        for variable, data in data_dict.items():
            output_folder = os.path.join(directory_path, variable)
            os.makedirs(output_folder, exist_ok=True)
            output_file = os.path.join(output_folder, f"{variable}{station_name}.csv")
            data.to_csv(output_file, index_label='time')
            
    def _process_variable(self, file_list, variable):
        dataframes = []
        for file_name in file_list:
            file_path = os.path.join(self._get_folder_path(variable), file_name)
            if os.path.isfile(file_path):
                with xr.open_mfdataset(file_path, combine='by_coords') as ds:
                    model_files = ds.isel(time=slice(None, -1))
                    lon_target, lat_target = np.meshgrid(self.stn_lon, self.stn_lat)
                    lat_source, lon_source = model_files.lat, model_files.lon
                    xs, ys, zs = self.lon_lat_to_cartesian(lon_source.values, lat_source.values)
                    xt, yt, zt = self.lon_lat_to_cartesian(lon_target.flatten(), lat_target.flatten())
                    tree = cKDTree(np.column_stack((xs.flatten(), ys.flatten(), zs.flatten())))
                    dist, inds = tree.query(np.column_stack((xt, yt, zt)), k=self.n_neighbours)

                    num_layers = model_files[variable].shape[1]
                    #print(f"Interpolating {variable} with {num_layers} layers.")
                    df_layers = pd.DataFrame()

                    for layer in range(num_layers):
                        interp_vals = self.interpolate(
                            model_files.isel(lev=layer),  # Asegúrate de usar 'lev'
                            lon_target,
                            dist,
                            inds,
                            variable
                        )
                        if model_files[variable].attrs.get('units') == 'kgm-3':
                            x = np.array([interp_val[0][0] * 1e9 for interp_val in interp_vals]) # Convert kg/m3 to ug/m3
                        else:
                            x = np.array([interp_val[0][0] for interp_val in interp_vals])

                        df_layers[f"layer_{layer + 1}"] = x

                    # Manejo del índice temporal
                    times = model_files['time'].values
                    if len(times) != len(df_layers):
                        raise ValueError(
                            f"Time dimension mismatch: {len(times)} time steps vs {len(df_layers)} data rows."
                        )
                    df_layers.index = times
                    dataframes.append(df_layers)
            else:
                print(f"No se encontró el archivo '{file_name}'.")
        return pd.concat(dataframes, axis=0)


#cargando staciones 
with open('aeronet_stations.json') as f:
    stn = json.load(f)

#haciendo la tupla de las estaciones
stations = [(name, data['lat'], data['lon']) for name, data in stn.items()]

#dividimos las estaciones en partes iguales
stations_split = np.array_split(stations, size)
variables = vti.brown_carbon

expid = 'a88m'
start_date = date(2018, 1, 1)
end_date = date(2018, 1, 20)

# Cada proceso procesa su porción de estaciones
for station in stations_split[rank]:
    name, lat, lon = station
    reader = NetCDFVertInterp(expid, start_date, end_date, lat, lon)
    output_file = reader.save_data('/esarchive/scratch/hnavarro/InterpolatedVertData', name, variables)

comm.Barrier()  # Wait for all processes to reach this point
if rank == 0:
    print("All processes have finished.")

comm.Finalize()

# To run this script, use the following command:
# mpiexec -n 4 python vertIntMONARCH.py